# From Guilhermelawless: 
# https://github.com/guilhermelawless/nano-dpow/blob/master/server/dpow/nano_websocket.py
from aiohttp import log
import asyncio
import websockets
import rapidjson as json
import traceback
from pippin.network.work_client import WorkClient


def subscription(topic: str, ack: bool = False, options: dict = None):
    d = {"action": "subscribe", "topic": topic, "ack": ack}
    if options is not None:
        d["options"] = options
    return d

class WebsocketClient(object):

    def __init__(self, uri, callback):
        self.uri = uri
        self.arrival_cb = callback
        self.ws = None
        self.stop = False

    async def setup(self, silent=False):
        try:
            self.ws = await websockets.connect(self.uri)
            await self.ws.send(json.dumps(subscription("confirmation", ack=True)))
            await self.ws.recv()  # ack
            await self.ws.send(json.dumps(subscription("active_difficulty", ack=True)))
            await self.ws.recv()
        except Exception as e:
            if not silent:
                log.server_logger.critical("NANO WS: Error connecting to websocket server. Check your settings in ~/PippinData/config.yaml")
                log.server_logger.error(traceback.format_exc())
            raise

    async def close(self):
        self.stop = True
        await self.ws.wait_closed()

    async def reconnect_forever(self):
        log.server_logger.warn("NANO WS: Attempting websocket reconnection every 30 seconds...")
        while not self.stop:
            try:
                await self.setup(silent=True)
                log.server_logger.warn("NANO WS: Connected to websocket!")
                break
            except:
                log.server_logger.debug("NANO WS: Websocket reconnection failed")
                await asyncio.sleep(30)

    async def loop(self):
        while not self.stop:
            try:
                rec = json.loads(await self.ws.recv())
                topic = rec.get("topic", None)
                if topic and topic == "confirmation":
                    await self.arrival_cb(rec["message"])
                elif topic and topic == "active_difficulty":
                    if "network_current" in rec["message"]:
                        log.server_logger.debug(f"Active difficulty changed to {rec['message']['network_current']}")
                        WorkClient.instance().active_difficulty = rec["message"]["network_current"]
            except KeyboardInterrupt:
                break
            except websockets.exceptions.ConnectionClosed as e:
                log.server_logger.error(f"NANO WS: Connection closed to websocket. Code: {e.code} , reason: {e.reason}.")
                await self.reconnect_forever()
            except Exception as e:
                log.server_logger.critical(f"NANO WS: Unknown exception while handling getting a websocket message:\n{traceback.format_exc()}")
                await self.reconnect_forever()



from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor

import asyncio
import pippin.config as config
import functools
import nanopy

class NanoUtil(object):
    _instance = None

    def __init__(self):
        raise RuntimeError('Call instance() instead')

    @classmethod
    def instance(cls, max_work_processes: int = 1, max_sign_threads: int = 1) -> 'NanoUtil':
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
            work_processes = max(config.Config.instance().max_work_processes, 0)
            if work_processes > 0:
                cls.process_pool = ProcessPoolExecutor(max_workers=work_processes)
            else:
                cls.process_pool = None
            cls.thread_pool = ThreadPoolExecutor(max_workers=max(config.Config.instance().max_sign_threads, 1))
            cls.thread_pool
        return cls._instance

    @classmethod
    async def close(cls):
        if hasattr(cls, 'process_pool') and cls.process_pool is not None:
            cls.process_pool.shutdown()
        if hasattr(cls, 'thread_pool') and cls.thread_pool is not None:
            cls.thread_pool.shutdown()
        if cls._instance is not None:
            cls._instance = None

    async def work_generate(self, hash: str, difficulty: str = None) -> str:
        """Run work_generate in ProcessPool"""
        if self.process_pool is None:
            raise WorkDisabled()

        difficulty = difficulty if difficulty is not None else nanopy.work_difficulty
        result = await asyncio.get_event_loop().run_in_executor(
            self.process_pool,
            functools.partial(
                nanopy.work_generate,
                hash,
                difficulty
            )
        )
        return result

    async def sign_block(self, private_key: str, block: dict) -> str:
        """Sign block in ThreadPool"""
        # We want to sign blocks in a ThreadPool to prevent blocking our main event loop
        result = await asyncio.get_event_loop().run_in_executor(
            self.thread_pool,
            functools.partial(
                nanopy.sign,
                private_key,
                block
            )
        )
        return result

class WorkDisabled(Exception):
    pass


import secrets
import string

class RandomUtil(object):

    @staticmethod
    def generate_seed() -> str:
        """Generate a random seed and return it"""
        seed = "".join([secrets.choice(string.hexdigits) for i in range(64)]).upper()
        return seed


from tortoise.models import Model
from tortoise import fields

import rapidjson

class Block(Model):
    """A local storage of blocks we create, primarily for send blocks"""
    account  = fields.ForeignKeyField('db.Account', on_delete=fields.CASCADE, related_name='blocks', index=True, null=True)
    adhoc_account  = fields.ForeignKeyField('db.AdHocAccount', on_delete=fields.CASCADE, related_name='blocks', index=True, null=True)
    block_hash = fields.CharField(max_length=64, unique=True)
    block = fields.JSONField(encoder=rapidjson.dumps, decoder=rapidjson.loads)
    send_id = fields.CharField(max_length=64, null=True, index=True)
    subtype = fields.CharField(max_length=10)
    created_at = fields.DatetimeField(auto_now_add=True)

    class Meta:
        table = 'blocks'
        unique_together = ('account', 'send_id')


from aiohttp import log
import asyncio
import websockets
import rapidjson as json
import traceback
import nanopy
import nanolib

class ConnectionClosed(Exception):
    pass

class DpowClient(object):
    """Websocket client for DPoW & BoomPoW"""
    DPOW_SERVER = 'wss://dpow.nanocenter.org/service_ws/'
    BPOW_SERVER = 'wss://bpow.banano.cc/service_ws/'

    def __init__(self, dpow_user: str, dpow_key: str, work_futures: dict = {}, bpow: bool = False):
        """work_futures is a dict of id, futures to keep track of requests and responses"""
        self.bpow = bpow
        self.url = self.BPOW_SERVER if self.bpow else self.DPOW_SERVER
        self.ws: websockets.WebSocketClientProtocol = None
        self.stop = False
        self.dpow_user = dpow_user
        self.dpow_key = dpow_key
        self.work_futures = work_futures

    def adjust_difficulty(self, difficulty: str) -> str:
        """Ensure sane difficulty limits for DPoW/BPoW"""
        if nanolib.work.derive_work_multiplier(difficulty, base_difficulty=nanopy.work_difficulty) > 8:
            return nanolib.work.derive_work_difficulty(8, base_difficulty=nanopy.work_difficulty)
        return difficulty

    async def request_work(self, id: str, hash: str, difficulty: str):
        """Request work from DPoW/BPoW WS"""
        if self.stop or self.ws is None:
            return
        try:
            if self.ws.closed:
                raise ConnectionClosed()
            req = {
                "user": self.dpow_user,
                "api_key": self.dpow_key,
                "hash": hash,
                "id": id,
                "timeout": 15,
                "difficulty": self.adjust_difficulty(difficulty)
            }
            await self.ws.send(json.dumps(req))
        except Exception:
            raise ConnectionClosed()        

    async def setup(self):
        try:
            self.ws = await websockets.connect(self.url)
            log.server_logger.info(f"Connected to {'BoomPoW' if self.bpow else 'Distributed PoW'} Service")
        except Exception as e:
            log.server_logger.critical("DPOW WS: Error connecting to websocket server.")
            log.server_logger.error(traceback.format_exc())
            raise

    async def close(self):
        self.stop = True
        await self.ws.wait_closed()

    async def reconnect_forever(self):
        log.server_logger.warn("DPOW WS: Attempting websocket reconnection every 30 seconds...")
        while not self.stop:
            try:
                await self.setup()
                log.server_logger.warn("DPOW WS: Connected to websocket!")
                break
            except:
                log.server_logger.debug("DPOW WS: Websocket reconnection failed")
                await asyncio.sleep(30)

    async def loop(self):
        while not self.stop:
            try:
                rec = json.loads(await self.ws.recv())
                request_id = rec.get("id", None)
                if request_id is not None:
                    try:
                        result = self.work_futures[str(request_id)]
                        if not result.done():
                            result.set_result(rec)
                    except KeyError:
                        pass                    
            except KeyboardInterrupt:
                break
            except websockets.exceptions.ConnectionClosed as e:
                log.server_logger.error(f"DPOW WS: Connection closed to websocket. Code: {e.code} , reason: {e.reason}.")
                await self.reconnect_forever()
            except Exception as e:
                log.server_logger.critical(f"DPOW WS: Unknown exception while handling getting a websocket message:\n{traceback.format_exc()}")
                await self.reconnect_forever()
